@@ -1,5 +1,5 @@
 /*
- * SPDX-FileCopyrightText: Copyright (c) 2019-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
+ * SPDX-FileCopyrightText: Copyright (c) 2019-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
  * SPDX-License-Identifier: MIT
  *
  * Permission is hereby granted, free of charge, to any person obtaining a
@@ -24,13 +24,14 @@
 #include "nv-pci-table.h"
 #include "nv-pci-types.h"
 #include "nv-pci.h"
-#include "nv-ibmnpu.h"
 #include "nv-msi.h"
 #include "nv-hypervisor.h"
+#include "nv-reg.h"
 
 #if defined(NV_VGPU_KVM_BUILD)
 #include "nv-vgpu-vfio-interface.h"
 #endif
+#include <linux/iommu.h>
 
 #if defined(NV_SEQ_READ_ITER_PRESENT)
 #include <linux/seq_file.h>
@@ -46,6 +47,8 @@
 #include <linux/pci-ats.h>
 #endif
 
+extern int NVreg_GrdmaPciTopoCheckOverride;
+
 static void
 nv_check_and_exclude_gpu(
     nvidia_stack_t *sp,
@@ -150,6 +153,20 @@ static void nv_get_pci_sysfs_config
 #endif
 }
 
+static int nv_resize_pcie_bars(struct pci_dev *pci_dev) {
+    nv_printf(NV_DBG_INFO, "NVRM: Resizable BAR is not supported on this driver version.\n");
+    return 0;
+}
+
+static void
+nv_init_coherent_link_info
+(
+    nv_state_t *nv
+)
+{
+    return;
+}
+
 static void nv_init_dynamic_power_management
 (
     nvidia_stack_t *sp,
@@ -173,21 +190,63 @@ static void nv_init_dynamic_power_management
         pr3_acpi_method_present = nv_acpi_power_resource_method_present(pci_dev->bus->self);
     }
 
+    // Support dynamic power management if device is a tegra PCI iGPU
+    rm_init_tegra_dynamic_power_management(sp, nv);
+
     rm_init_dynamic_power_management(sp, nv, pr3_acpi_method_present);
 }
 
-static int nv_resize_pcie_bars(struct pci_dev *pci_dev) {
-    nv_printf(NV_DBG_INFO, "NVRM: Resizable BAR is not supported on this driver version.\n");
-    return 0;
-}
-
-static void
-nv_init_coherent_link_info
-(
-    nv_state_t *nv
-)
+static NvBool
+nv_pci_validate_bars(const struct pci_dev *pci_dev, NvBool only_bar0)
 {
-    return;
+    unsigned int i, j;
+    NvBool last_bar_64bit = NV_FALSE;
+
+    for (i = 0, j = 0; i < NVRM_PCICFG_NUM_BARS && j < NV_GPU_NUM_BARS; i++)
+    {
+        if (NV_PCI_RESOURCE_VALID(pci_dev, i))
+        {
+            if ((NV_PCI_RESOURCE_FLAGS(pci_dev, i) & PCI_BASE_ADDRESS_MEM_TYPE_64) &&
+                (NV_PCI_RESOURCE_FLAGS(pci_dev, i) & PCI_BASE_ADDRESS_MEM_PREFETCH))
+            {
+                last_bar_64bit = NV_TRUE;
+            }
+
+            //
+            // If we are here, then we have found a valid BAR -- 32 or 64-bit.
+            //
+            j++;
+
+            if (only_bar0)
+                return NV_TRUE;
+
+            continue;
+        }
+
+        //
+        // If last_bar_64bit is "true" then, we are looking at the 2nd (upper)
+        // half of the 64-bit BAR. This is typically all 0s which looks invalid
+        // but it's normal and not a problem and we can ignore it and continue.
+        //
+        if (last_bar_64bit)
+        {
+            last_bar_64bit = NV_FALSE;
+            continue;
+        }
+
+        // Invalid 32 or 64-bit BAR.
+        nv_printf(NV_DBG_ERRORS,
+            "NVRM: This PCI I/O region assigned to your NVIDIA device is invalid:\n"
+            "NVRM: BAR%d is %" NvU64_fmtu "M @ 0x%" NvU64_fmtx " (PCI:%04x:%02x:%02x.%x)\n", i,
+            (NvU64)(NV_PCI_RESOURCE_SIZE(pci_dev, i) >> 20),
+            (NvU64)NV_PCI_RESOURCE_START(pci_dev, i),
+            NV_PCI_DOMAIN_NUMBER(pci_dev), NV_PCI_BUS_NUMBER(pci_dev),
+            NV_PCI_SLOT_NUMBER(pci_dev), PCI_FUNC(pci_dev->devfn));
+
+        return NV_FALSE;
+    }
+
+    return NV_TRUE;
 }
 
 /* find nvidia devices and set initial state */
@@ -205,9 +264,9 @@ nv_pci_probe
     nvidia_stack_t *sp = NULL;
     NvBool prev_nv_ats_supported = nv_ats_supported;
     NV_STATUS status;
-    NvBool last_bar_64bit = NV_FALSE;
     NvU8 regs_bar_index = nv_bar_index_to_os_bar_index(pci_dev,
                                                        NV_GPU_BAR_INDEX_REGS);
+    NvBool bar0_requested = NV_FALSE;
 
     nv_printf(NV_DBG_SETUP, "NVRM: probing 0x%x 0x%x, class 0x%x\n",
         pci_dev->vendor, pci_dev->device, pci_dev->class);
@@ -281,103 +340,9 @@ nv_pci_probe
         goto failed;
     }
 
-    for (i = 0, j = 0; i < NVRM_PCICFG_NUM_BARS && j < NV_GPU_NUM_BARS; i++)
-    {
-        if (NV_PCI_RESOURCE_VALID(pci_dev, i))
-        {
-#if defined(NV_PCI_MAX_MMIO_BITS_SUPPORTED)
-            if ((NV_PCI_RESOURCE_FLAGS(pci_dev, i) & PCI_BASE_ADDRESS_MEM_TYPE_64) &&
-                ((NV_PCI_RESOURCE_START(pci_dev, i) >> NV_PCI_MAX_MMIO_BITS_SUPPORTED)))
-            {
-                nv_printf(NV_DBG_ERRORS,
-                    "NVRM: This is a 64-bit BAR mapped above %dGB by the system\n"
-                    "NVRM: BIOS or the %s kernel. This PCI I/O region assigned\n"
-                    "NVRM: to your NVIDIA device is not supported by the kernel.\n"
-                    "NVRM: BAR%d is %dM @ 0x%llx (PCI:%04x:%02x:%02x.%x)\n",
-                    (1 << (NV_PCI_MAX_MMIO_BITS_SUPPORTED - 30)),
-                    NV_KERNEL_NAME, i,
-                    (NV_PCI_RESOURCE_SIZE(pci_dev, i) >> 20),
-                    (NvU64)NV_PCI_RESOURCE_START(pci_dev, i),
-                    NV_PCI_DOMAIN_NUMBER(pci_dev),
-                    NV_PCI_BUS_NUMBER(pci_dev), NV_PCI_SLOT_NUMBER(pci_dev),
-                    PCI_FUNC(pci_dev->devfn));
-                goto failed;
-            }
-#endif
-            if ((NV_PCI_RESOURCE_FLAGS(pci_dev, i) & PCI_BASE_ADDRESS_MEM_TYPE_64) &&
-                (NV_PCI_RESOURCE_FLAGS(pci_dev, i) & PCI_BASE_ADDRESS_MEM_PREFETCH))
-            {
-                struct pci_dev *bridge = pci_dev->bus->self;
-                NvU32 base_upper, limit_upper;
-
-                last_bar_64bit = NV_TRUE;
-
-                if (bridge == NULL)
-                    goto next_bar;
-
-                pci_read_config_dword(pci_dev, NVRM_PCICFG_BAR_OFFSET(i) + 4,
-                                      &base_upper);
-                if (base_upper == 0)
-                    goto next_bar;
-
-                pci_read_config_dword(bridge, PCI_PREF_BASE_UPPER32,
-                        &base_upper);
-                pci_read_config_dword(bridge, PCI_PREF_LIMIT_UPPER32,
-                        &limit_upper);
-
-                if ((base_upper != 0) && (limit_upper != 0))
-                    goto next_bar;
-
-                nv_printf(NV_DBG_ERRORS,
-                    "NVRM: This is a 64-bit BAR mapped above 4GB by the system\n"
-                    "NVRM: BIOS or the %s kernel, but the PCI bridge\n"
-                    "NVRM: immediately upstream of this GPU does not define\n"
-                    "NVRM: a matching prefetchable memory window.\n",
-                    NV_KERNEL_NAME);
-                nv_printf(NV_DBG_ERRORS,
-                    "NVRM: This may be due to a known Linux kernel bug.  Please\n"
-                    "NVRM: see the README section on 64-bit BARs for additional\n"
-                    "NVRM: information.\n");
-                goto failed;
-            }
-
-next_bar:
-            //
-            // If we are here, then we have found a valid BAR -- 32 or 64-bit.
-            //
-            j++;
-            continue;
-        }
-
-        //
-        // If last_bar_64bit is "true" then, we are looking at the 2nd (upper)
-        // half of the 64-bit BAR. This is typically all 0s which looks invalid
-        // but it's normal and not a problem and we can ignore it and continue.
-        //
-        if (last_bar_64bit)
-        {
-            last_bar_64bit = NV_FALSE;
-            continue;
-        }
-
-        // Invalid 32 or 64-bit BAR.
-        nv_printf(NV_DBG_ERRORS,
-            "NVRM: This PCI I/O region assigned to your NVIDIA device is invalid:\n"
-            "NVRM: BAR%d is %" NvU64_fmtu "M @ 0x%" NvU64_fmtx " (PCI:%04x:%02x:%02x.%x)\n", i,
-            (NvU64)(NV_PCI_RESOURCE_SIZE(pci_dev, i) >> 20),
-            (NvU64)NV_PCI_RESOURCE_START(pci_dev, i),
-            NV_PCI_DOMAIN_NUMBER(pci_dev), NV_PCI_BUS_NUMBER(pci_dev),
-            NV_PCI_SLOT_NUMBER(pci_dev), PCI_FUNC(pci_dev->devfn));
-
-        // With GH180 C2C, VF BAR1/2 are disabled and therefore expected to be 0.
-        if (j != NV_GPU_BAR_INDEX_REGS)
-        {
-            nv_printf(NV_DBG_INFO, "NVRM: ignore invalid BAR failure for BAR%d\n", j);
-            continue;
-        }
-
+    // Validate if BAR0 is usable
+    if (!nv_pci_validate_bars(pci_dev, /* only_bar0 = */ NV_TRUE))
         goto failed;
-    }
 
     if (!request_mem_region(NV_PCI_RESOURCE_START(pci_dev, regs_bar_index),
                             NV_PCI_RESOURCE_SIZE(pci_dev, regs_bar_index),
@@ -392,11 +357,7 @@ next_bar:
         goto failed;
     }
 
-    if (nv_resize_pcie_bars(pci_dev)) {
-        nv_printf(NV_DBG_ERRORS,
-            "NVRM: Fatal Error while attempting to resize PCIe BARs.\n");
-        goto failed;
-    }
+    bar0_requested = NV_TRUE;
 
     NV_KZALLOC(nvl, sizeof(nv_linux_state_t));
     if (nvl == NULL)
@@ -407,6 +368,21 @@ next_bar:
 
     nv  = NV_STATE_PTR(nvl);
 
+    for (i = 0; i < NVRM_PCICFG_NUM_BARS; i++)
+    {
+        if ((NV_PCI_RESOURCE_VALID(pci_dev, i)) &&
+            (NV_PCI_RESOURCE_FLAGS(pci_dev, i) & PCI_BASE_ADDRESS_SPACE)
+                == PCI_BASE_ADDRESS_SPACE_MEMORY)
+        {
+            nv->bars[NV_GPU_BAR_INDEX_REGS].offset = NVRM_PCICFG_BAR_OFFSET(i);
+            nv->bars[NV_GPU_BAR_INDEX_REGS].cpu_address = NV_PCI_RESOURCE_START(pci_dev, i);
+            nv->bars[NV_GPU_BAR_INDEX_REGS].size = NV_PCI_RESOURCE_SIZE(pci_dev, i);
+
+            break;
+        }
+    }
+    nv->regs = &nv->bars[NV_GPU_BAR_INDEX_REGS];
+
     pci_set_drvdata(pci_dev, (void *)nvl);
 
     /* default to 32-bit PCI bus address space */
@@ -433,55 +409,72 @@ next_bar:
         goto err_not_supported;
     }
 
+    if ((rm_is_supported_device(sp, nv)) != NV_OK)
+        goto err_not_supported;
+
+    // Wire RM HAL
+    if (!rm_init_private_state(sp, nv))
+    {
+        NV_DEV_PRINTF(NV_DBG_ERRORS, nv, "rm_init_private_state() failed!\n");
+        goto err_zero_dev;
+    }
+
+    if (!nv->is_tegra_pci_igpu &&
+        !pci_devid_is_self_hosted(pci_dev->device) &&
+        !nv_pci_validate_bars(pci_dev, /* only_bar0 = */ NV_FALSE))
+        goto err_zero_dev;
+
+    if (nv_resize_pcie_bars(pci_dev)) {
+        nv_printf(NV_DBG_ERRORS,
+            "NVRM: Fatal Error while attempting to resize PCIe BARs.\n");
+        goto err_zero_dev;
+    }
+
     nvl->all_mappings_revoked = NV_TRUE;
     nvl->safe_to_mmap = NV_TRUE;
     nvl->gpu_wakeup_callback_needed = NV_TRUE;
     INIT_LIST_HEAD(&nvl->open_files);
 
-    for (i = 0, j = 0; i < NVRM_PCICFG_NUM_BARS && j < NV_GPU_NUM_BARS; i++)
+    if (!nv->is_tegra_pci_igpu)
     {
-        if ((NV_PCI_RESOURCE_VALID(pci_dev, i)) &&
-            (NV_PCI_RESOURCE_FLAGS(pci_dev, i) & PCI_BASE_ADDRESS_SPACE)
-                == PCI_BASE_ADDRESS_SPACE_MEMORY)
+        for (i = 0, j = 0; i < NVRM_PCICFG_NUM_BARS && j < NV_GPU_NUM_BARS; i++)
         {
-            nv->bars[j].offset = NVRM_PCICFG_BAR_OFFSET(i);
-            nv->bars[j].cpu_address = NV_PCI_RESOURCE_START(pci_dev, i);
-            nv->bars[j].size = NV_PCI_RESOURCE_SIZE(pci_dev, i);
-            j++;
+            if (j == NV_GPU_BAR_INDEX_REGS)
+            {
+                j++;
+                continue;
+            }
+
+            if ((NV_PCI_RESOURCE_VALID(pci_dev, i)) &&
+                (NV_PCI_RESOURCE_FLAGS(pci_dev, i) & PCI_BASE_ADDRESS_SPACE)
+                    == PCI_BASE_ADDRESS_SPACE_MEMORY)
+            {
+                nv->bars[j].offset = NVRM_PCICFG_BAR_OFFSET(i);
+                nv->bars[j].cpu_address = NV_PCI_RESOURCE_START(pci_dev, i);
+                nv->bars[j].size = NV_PCI_RESOURCE_SIZE(pci_dev, i);
+                j++;
+            }
         }
     }
-    nv->regs = &nv->bars[NV_GPU_BAR_INDEX_REGS];
-    nv->fb   = &nv->bars[NV_GPU_BAR_INDEX_FB];
 
+    nv->fb   = &nv->bars[NV_GPU_BAR_INDEX_FB];
     nv->interrupt_line = pci_dev->irq;
 
     NV_ATOMIC_SET(nvl->numa_info.status, NV_IOCTL_NUMA_STATUS_DISABLED);
     nvl->numa_info.node_id = NUMA_NO_NODE;
 
-    nv_init_ibmnpu_info(nv);
-
-    nv_init_coherent_link_info(nv);
-
-#if defined(NVCPU_PPC64LE)
-    // Use HW NUMA support as a proxy for ATS support. This is true in the only
-    // PPC64LE platform where ATS is currently supported (IBM P9).
-    nv_ats_supported &= nv_platform_supports_numa(nvl);
-#else
-    nv_ats_supported = NV_FALSE;
-#endif
-    if (nv_ats_supported)
+    if (pci_devid_is_self_hosted(pci_dev->device))
     {
-        NV_DEV_PRINTF(NV_DBG_INFO, nv, "ATS supported by this GPU!\n");
-    }
-    else
-    {
-        NV_DEV_PRINTF(NV_DBG_INFO, nv, "ATS not supported by this GPU. "
-                      "Disabling ATS support for all the GPUs in the system!\n");
+        nv_init_coherent_link_info(nv);
     }
 
+    nv_ats_supported |= nv->ats_support;
+
+    nv_clk_get_handles(nv);
+
     pci_set_master(pci_dev);
 
-#if defined(CONFIG_VGA_ARB) && !defined(NVCPU_PPC64LE)
+#if defined(CONFIG_VGA_ARB)
 #if defined(VGA_DEFAULT_DEVICE)
 #if defined(NV_VGA_TRYGET_PRESENT)
     vga_tryget(VGA_DEFAULT_DEVICE, VGA_RSRC_LEGACY_MASK);
@@ -494,16 +487,7 @@ next_bar:
     if (status == NV_ERR_GPU_IS_LOST)
     {
         NV_DEV_PRINTF(NV_DBG_INFO, nv, "GPU is lost, skipping nv_pci_probe\n");
-        goto err_not_supported;
-    }
-
-    if ((rm_is_supported_device(sp, nv)) != NV_OK)
-        goto err_not_supported;
-
-    if (!rm_init_private_state(sp, nv))
-    {
-        NV_DEV_PRINTF(NV_DBG_ERRORS, nv, "rm_init_private_state() failed!\n");
-        goto err_zero_dev;
+        goto err_gpu_lost;
     }
 
     nv->cpu_numa_node_id = dev_to_node(nvl->dev);
@@ -511,7 +495,7 @@ next_bar:
     if (nv_linux_init_open_q(nvl) != 0)
     {
         NV_DEV_PRINTF(NV_DBG_ERRORS, nv, "nv_linux_init_open_q() failed!\n");
-        goto err_zero_dev;
+        goto err_gpu_lost;
     }
 
     nv_printf(NV_DBG_INFO,
@@ -547,6 +531,7 @@ next_bar:
 
     nv_init_dynamic_power_management(sp, pci_dev);
 
+
     nv_procfs_add_gpu(nvl);
 
     /* Parse and set any per-GPU registry keys specified. */
@@ -558,7 +543,7 @@ next_bar:
     if (nvidia_vgpu_vfio_probe(nvl->pci_dev) != NV_OK)
     {
         NV_DEV_PRINTF(NV_DBG_ERRORS, nv, "Failed to register device to vGPU VFIO module");
-        goto err_vgpu_kvm;
+        goto err_free_all;
     }
 #endif
 
@@ -577,13 +562,18 @@ next_bar:
      * after enabling dynamic power management.
      */
     rm_enable_dynamic_power_management(sp, nv);
+
+    /*
+     * This must be the last action in nv_pci_probe(). Do not add code after this line.
+     */
+    rm_notify_gpu_addition(sp, nv);
+
     nv_kmem_cache_free_stack(sp);
 
     return 0;
 
-#if defined(NV_VGPU_KVM_BUILD)
-err_vgpu_kvm:
-#endif
+goto err_free_all;
+err_free_all:
     nv_procfs_remove_gpu(nvl);
     rm_cleanup_dynamic_power_management(sp, nv);
     pm_vt_switch_unregister(nvl->dev);
@@ -592,21 +582,25 @@ err_vgpu_kvm:
     UNLOCK_NV_LINUX_DEVICES();
 err_add_device:
     nv_linux_stop_open_q(nvl);
+err_gpu_lost:
+    nv_clk_clear_handles(nv);
+    nv_ats_supported = prev_nv_ats_supported;
 err_zero_dev:
     rm_free_private_state(sp, nv);
 err_not_supported:
-    nv_ats_supported = prev_nv_ats_supported;
-    nv_destroy_ibmnpu_info(nv);
     nv_lock_destroy_locks(sp, nv);
+failed:
+    if (bar0_requested)
+    {
+        release_mem_region(NV_PCI_RESOURCE_START(pci_dev, regs_bar_index),
+                           NV_PCI_RESOURCE_SIZE(pci_dev, regs_bar_index));
+    }
+    NV_PCI_DISABLE_DEVICE(pci_dev);
+    pci_set_drvdata(pci_dev, NULL);
     if (nvl != NULL)
     {
         NV_KFREE(nvl, sizeof(nv_linux_state_t));
     }
-    release_mem_region(NV_PCI_RESOURCE_START(pci_dev, regs_bar_index),
-                       NV_PCI_RESOURCE_SIZE(pci_dev, regs_bar_index));
-    NV_PCI_DISABLE_DEVICE(pci_dev);
-    pci_set_drvdata(pci_dev, NULL);
-failed:
     nv_kmem_cache_free_stack(sp);
     return -1;
 }
@@ -657,6 +651,9 @@ nv_pci_remove(struct pci_dev *pci_dev)
 
     LOCK_NV_LINUX_DEVICES();
     down(&nvl->ldata_lock);
+    nv->flags |= NV_FLAG_PCI_REMOVE_IN_PROGRESS;
+
+    rm_notify_gpu_removal(sp, nv);
 
     /*
      * Sanity check: A removed device shouldn't have a non-zero usage_count.
@@ -716,6 +713,8 @@ nv_pci_remove(struct pci_dev *pci_dev)
     /* Remove proc entry for this GPU */
     nv_procfs_remove_gpu(nvl);
 
+    nv_clk_clear_handles(nv);
+
     rm_cleanup_dynamic_power_management(sp, nv);
 
     nv->removed = NV_TRUE;
@@ -746,9 +745,6 @@ nv_pci_remove(struct pci_dev *pci_dev)
         nvl->sysfs_config_file = NULL;
     }
 
-    nv_unregister_ibmnpu_devices(nv);
-    nv_destroy_ibmnpu_info(nv);
-
     if (NV_ATOMIC_READ(nvl->usage_count) == 0)
     {
         nv_lock_destroy_locks(sp, nv);
@@ -964,96 +960,98 @@ nv_pci_count_devices(void)
     return count;
 }
 
-#if defined(NV_PCI_ERROR_RECOVERY)
-static pci_ers_result_t
-nv_pci_error_detected(
-    struct pci_dev *pci_dev,
-    nv_pci_channel_state_t error
+/*
+ * On coherent platforms that support BAR1 mappings for GPUDirect RDMA,
+ * dma-buf and nv-p2p subsystems need to ensure the 2 devices belong to
+ * the same IOMMU group.
+ */
+NvBool nv_pci_is_valid_topology_for_direct_pci(
+    nv_state_t     *nv,
+    struct pci_dev *peer
 )
 {
-    nv_linux_state_t *nvl = pci_get_drvdata(pci_dev);
+    struct pci_dev *pdev0 = to_pci_dev(nv->dma_dev->dev);
+    struct pci_dev *pdev1 = peer;
 
-    if ((nvl == NULL) || (nvl->pci_dev != pci_dev))
+    if (!nv->coherent)
     {
-        nv_printf(NV_DBG_ERRORS, "NVRM: %s: invalid device!\n", __FUNCTION__);
-        return PCI_ERS_RESULT_NONE;
+        return NV_FALSE;
     }
 
-    /*
-     * Tell Linux to continue recovery of the device. The kernel will enable
-     * MMIO for the GPU and call the mmio_enabled callback.
-     */
-    return PCI_ERS_RESULT_CAN_RECOVER;
+    switch (NVreg_GrdmaPciTopoCheckOverride) {
+        case NV_REG_GRDMA_PCI_TOPO_CHECK_OVERRIDE_ALLOW_ACCESS:
+            return NV_TRUE;
+        case NV_REG_GRDMA_PCI_TOPO_CHECK_OVERRIDE_DENY_ACCESS:
+            return NV_FALSE;
+        default:
+           return (pdev0->dev.iommu_group == pdev1->dev.iommu_group);
+    }
 }
 
-static pci_ers_result_t
-nv_pci_mmio_enabled(
-    struct pci_dev *pci_dev
+NvBool nv_pci_has_common_pci_switch(
+    nv_state_t     *nv,
+    struct pci_dev *peer
 )
 {
-    NV_STATUS         status = NV_OK;
-    nv_stack_t       *sp = NULL;
-    nv_linux_state_t *nvl = pci_get_drvdata(pci_dev);
-    nv_state_t       *nv = NULL;
+    struct pci_dev *pci_dev0, *pci_dev1;
 
-    if ((nvl == NULL) || (nvl->pci_dev != pci_dev))
+    pci_dev0 = pci_upstream_bridge(to_pci_dev(nv->dma_dev->dev));
+
+    while (pci_dev0 != NULL)
     {
-        nv_printf(NV_DBG_ERRORS, "NVRM: %s: invalid device!\n", __FUNCTION__);
-        goto done;
+        pci_dev1 = pci_upstream_bridge(peer);
+
+        while (pci_dev1 != NULL)
+        {
+            if (pci_dev0 == pci_dev1)
+                return NV_TRUE;
+
+            pci_dev1 = pci_upstream_bridge(pci_dev1);
+        }
+
+        pci_dev0 = pci_upstream_bridge(pci_dev0);
     }
 
-    nv = NV_STATE_PTR(nvl);
-
-    if (nv_kmem_cache_alloc_stack(&sp) != 0)
-    {
-        nv_printf(NV_DBG_ERRORS, "NVRM: %s: failed to allocate stack!\n",
-            __FUNCTION__);
-        goto done;
-    }
-
-    NV_DEV_PRINTF(NV_DBG_ERRORS, nv, "A fatal error was detected.\n");
-
-    /*
-     * MMIO should be re-enabled now. If we still get bad reads, there's
-     * likely something wrong with the adapter itself that will require a
-     * reset. This should let us know whether the GPU has completely fallen
-     * off the bus or just did something the host didn't like.
-     */
-    status = rm_is_supported_device(sp, nv);
-    if (status != NV_OK)
-    {
-        NV_DEV_PRINTF(NV_DBG_ERRORS, nv,
-            "The kernel has enabled MMIO for the device,\n"
-            "NVRM: but it still appears unreachable. The device\n"
-            "NVRM: will not function properly until it is reset.\n");
-    }
-
-    status = rm_log_gpu_crash(sp, nv);
-    if (status != NV_OK)
-    {
-        NV_DEV_PRINTF_STATUS(NV_DBG_ERRORS, nv, status,
-                      "Failed to log crash data\n");
-        goto done;
-    }
-
-done:
-    if (sp != NULL)
-    {
-        nv_kmem_cache_free_stack(sp);
-    }
-
-    /*
-     * Tell Linux to abandon recovery of the device. The kernel might be able
-     * to recover the device, but RM and clients don't yet support that.
-     */
-    return PCI_ERS_RESULT_DISCONNECT;
+    return NV_FALSE;
 }
 
-struct pci_error_handlers nv_pci_error_handlers = {
-    .error_detected = nv_pci_error_detected,
-    .mmio_enabled   = nv_pci_mmio_enabled,
-};
-#endif
+NvBool NV_API_CALL nv_grdma_pci_topology_supported(
+    nv_state_t      *nv,
+    nv_dma_device_t *dma_peer
+)
+{
+    //
+    // Skip topo check on coherent platforms since
+    // NIC can map over C2C anyway and PCIe topology shouldn't matter.
+    //
+    if (nv->coherent)
+    {
+        return NV_TRUE;
+    }
+
+    switch (NVreg_GrdmaPciTopoCheckOverride)
+    {
+        case NV_REG_GRDMA_PCI_TOPO_CHECK_OVERRIDE_ALLOW_ACCESS:
+            return NV_TRUE;
+        case NV_REG_GRDMA_PCI_TOPO_CHECK_OVERRIDE_DENY_ACCESS:
+            return NV_FALSE;
+        default:
+            break;
+    }
+
+    // Allow RDMA by default on passthrough VMs.
+    if ((nv->flags & NV_FLAG_PASSTHRU) != 0)
+        return NV_TRUE;
+
+    //
+    // Only allow RDMA on unsupported chipsets if there exists
+    // a common PCI switch between the GPU and the other device
+    //
+    if ((nv->flags & NV_FLAG_PCI_P2P_UNSUPPORTED_CHIPSET) != 0)
+        return nv_pci_has_common_pci_switch(nv, to_pci_dev(dma_peer->dev));
+
+    return NV_TRUE;
+}
 
 #if defined(CONFIG_PM)
 extern struct dev_pm_ops nv_pm_ops;
@@ -1072,9 +1070,6 @@ struct pci_driver nv_pci_driver = {
 #if defined(CONFIG_PM)
     .driver.pm = &nv_pm_ops,
 #endif
-#if defined(NV_PCI_ERROR_RECOVERY)
-    .err_handler = &nv_pci_error_handlers,
-#endif
 };
 
 void nv_pci_unregister_driver(void)
